"""
Debugging Module
"""

import os
import sys
import logging

from typing import (
    Any,
    Callable,
    Dict,
    Tuple,
    TypeVar,
    Union,
    Optional,
    cast,
)

# set the level of the root logger
_root = logging.getLogger()  # pylint: disable=invalid-name
_LOG_LEVEL = "WARNING"
if os.environ.get("LOG"):
    _LOG_LEVEL = os.environ["LOG"].upper()
_NUMERIC_LEVEL = getattr(logging, _LOG_LEVEL, None)
if not isinstance(_NUMERIC_LEVEL, int):
    raise ValueError("Invalid log level: %s" % _LOG_LEVEL)
_root.setLevel(_NUMERIC_LEVEL)

# add a stream handler for warnings and up
hdlr = logging.StreamHandler(stream=sys.stderr)
hdlr.setLevel(_NUMERIC_LEVEL)
hdlr.setFormatter(logging.Formatter("%(levelname)s:%(name)s %(message)s"))
_root.addHandler(hdlr)
_root.debug("log level %s", _LOG_LEVEL)
del hdlr

# module level logger
_log = logging.getLogger(__name__)  # pylint: disable=invalid-name

# types
FuncType = Callable[..., Any]
FuncTypeVar = TypeVar("FuncTypeVar", bound=FuncType)


def debugging(obj: FuncTypeVar) -> FuncTypeVar:
    """
    Decorator function that attaches a debugging logger to a class or function.
    """
    # create a logger for this object
    logger = logging.getLogger(obj.__module__ + "." + obj.__qualname__)

    # make it available to instances
    setattr(obj, "log_logger", logger)
    setattr(obj, "log_debug", logger.debug)

    return obj


# pylint: disable=bad-continuation
@debugging
def create_log_handler(
    logger: Union[str, logging.Logger] = "",
    handler: Optional[logging.StreamHandler] = None,
    level: int = logging.DEBUG,
) -> None:
    """
    Add a stream handler with our custom formatter to a logger.
    """
    logger_ref: logging.Logger

    if isinstance(logger, logging.Logger):
        logger_ref = logger

    elif isinstance(logger, str):
        # check for root
        if not logger:
            logger_ref = _log

        # check for a valid logger name
        elif logger not in logging.Logger.manager.loggerDict:  # type: ignore
            raise RuntimeError("not a valid logger name: %r" % (logger,))

        # get the logger
        logger_ref = logging.getLogger(logger)

    else:
        raise RuntimeError("not a valid logger reference: %r" % (logger,))

    # make a handler if one wasn't provided
    if not handler:
        handler = logging.StreamHandler()
        handler.setLevel(level)

        # use our formatter
        handler.setFormatter(logging.Formatter("%(levelname)s:%(name)s %(message)s"))

    # add it to the logger
    logger_ref.addHandler(handler)

    # make sure the logger has at least this level
    logger_ref.setLevel(level)


class LoggingMetaclass(type):
    """
    Logging Metaclass
    """

    def __new__(
        cls: Any,
        clsname: str,
        superclasses: Tuple[type, ...],
        attributedict: Dict[str, Any],
    ) -> "LoggingMetaclass":
        """
        Called for new classes inheriting from Logging, this calls the
        debugging function to set the _logger and _debug references to
        a Logger and its debug() method.
        """
        return cast(
            LoggingMetaclass,
            debugging(type.__new__(cls, clsname, superclasses, attributedict)),
        )


# pylint: disable=too-few-public-methods
class Logging(metaclass=LoggingMetaclass):
    """
    Inherit from this class to get a log_logger and log_debug method.
    """

    log_logger: logging.Logger
    log_debug: Callable[..., None]



"""
Consumer for RedisMQ
"""
from __future__ import annotations

import asyncio
import json

from typing import TYPE_CHECKING, Any, Dict, Callable

from .debugging import debugging

if TYPE_CHECKING:
    from .client import Client


@debugging
class Consumer:  # pylint: disable=too-few-public-methods
    """
    Consumes messages
    """

    client: Client
    stream_name: str
    group_name: str
    consumer_name: str
    min_idle_time: int

    log_debug: Callable[..., None]

    def __init__(
        self,
        client: Client,
        stream_name: str,
        group_name: str,
        consumer_name: str,
        scan_pending_on_start: bool = True,
        claim_stale_messages: bool = True,
        min_idle_time: int = 60000,
    ) -> None:
        """
        default constructor
        """
        self.client = client
        self.stream_name = stream_name
        self.group_name = group_name
        self.consumer_name = consumer_name
        self.scan_pending_on_start = scan_pending_on_start
        self.claim_stale_messages = claim_stale_messages
        self.min_idle_time = min_idle_time

        # by default just read new messages that haven't been delivered
        self.latest_id = b">"
        self.ready = True
        self.xread_timeout = 5

    # close()
    def close(self) -> None:
        """
        used to close a consumer gracefully if it is waiting to read
        """
        self.ready = False

    # to be used to cancel a read
    async def read(self) -> "Payload":
        """
        Read a message from the stream.
        """
        Consumer.log_debug("read")
        closed_payload = self.Payload(self, 0, {'message': '{"error": "closed"}'})
        if not self.ready:
            return closed_payload

        connection = await self.client.redis
        if self.scan_pending_on_start:
            (
                pending_count,
                min_id,
                max_id,
                pending_consumers,
            ) = await connection.xpending(self.stream_name, self.group_name)
            if pending_count:
                Consumer.log_debug(
                    f"    - pending summary {self.stream_name!r}: {pending_count!r}, \
                        {min_id!r}..{max_id!r}"
                )
                for pending_consumer, pending_consumer_count in pending_consumers:
                    if pending_consumer == self.consumer_name:
                        Consumer.log_debug("    - this consumer has pending messages")
                    else:
                        Consumer.log_debug(
                            f"    - pending messages for {pending_consumer!r}"
                        )

                    pending_messages = await connection.xpending(
                        self.stream_name,
                        self.group_name,
                        b"-",
                        b"+",
                        pending_consumer_count,
                        pending_consumer,
                    )
                    for (
                        msg_id,
                        consumer_name, # pylint: disable=unused-variable
                        idle_time,
                        delivered,
                    ) in pending_messages:
                        Consumer.log_debug(
                            f"        {msg_id!r}, idle {idle_time/1000.0!r}s, \
                                delivered {delivered!r}"
                        )
                        if self.claim_stale_messages:
                            retcode = await connection.xclaim(
                                self.stream_name,
                                self.group_name,
                                self.consumer_name,
                                self.min_idle_time,
                                msg_id,
                            )
                            Consumer.log_debug(f"        claim: {retcode!r}")

                            self.latest_id = b"0-0"
            else:
                Consumer.log_debug("    - no pending messages")

            # assume we'll get caught up
            self.scan_pending_on_start = False
        # only loop if ready, recheck periodically
        while self.ready:
            Consumer.log_debug("    - latest_id: %r", self.latest_id)

            args = {
                "group_name": self.group_name,
                "consumer_name": self.consumer_name,
                "streams": [self.stream_name],
                "timeout": self.xread_timeout,
                "count": 1,
                "latest_ids": [self.latest_id],
                "no_ack": False,
            }
            messages = await connection.xread_group(**args)
            Consumer.log_debug("    - messages: %r", messages)
            if messages:
                break

            # loop around again, read the next new message
            self.latest_id = b">"

        if not self.ready:
            return closed_payload
        (stream, msg_id, payload) = messages[0]
        payload_dict = dict(payload)
        Consumer.log_debug(
            "    - stream %s, id %s, payload %s", stream, msg_id, payload_dict
        )

        # assume this message is going to be processed, the next time read() is
        # called it will pick up the next claimed pending message, otherwise
        # loop around and read the next new message
        self.latest_id = msg_id

        return self.Payload(self, msg_id, payload_dict)

    class Payload:
        """
        Encapsulates the payload wrapped around a message and exposes an ack()
        function.
        """

        def __init__(
            self, consumer: "Consumer", msg_id: str, payload_dict: Dict[str, Any]
        ) -> None:
            Consumer.log_debug("__init__ %r %r", msg_id, payload_dict)

            self.consumer = consumer
            self.msg_id = msg_id
            self.response_channel = payload_dict.get("response_channel", None)
            try:
                self.message = json.loads(payload_dict["message"])
            except json.decoder.JSONDecodeError:
                Consumer.log_debug("    - unable to decode message, log this event")
                asyncio.ensure_future(
                    self.consumer.client.redis.xack(
                        self.consumer.stream_name, self.consumer.group_name, self.msg_id
                    )
                )

        async def ack(self, response: Any = None, error: Any = None) -> None:
            """
            Acks the message on the stream and publishes the response on the
            responseChannel, if provided.
            """
            Consumer.log_debug("ack %r %r", response, error)

            Consumer.log_debug("    - msg_id: %r", self.msg_id)
            with (await self.consumer.client.redis) as connection:
                await connection.xack(
                    self.consumer.stream_name, self.consumer.group_name, self.msg_id
                )
                if self.response_channel is not None:
                    Consumer.log_debug(
                        "    - response channel: %r", self.response_channel
                    )
                    m_response = {"message": response, "error": error}
                    await connection.publish_json(self.response_channel, m_response)



"""
Producer for RedisMQ
"""
from __future__ import annotations

import asyncio
import json
import time

from typing import TYPE_CHECKING, Any, Callable, Dict, cast

from .debugging import debugging

if TYPE_CHECKING:
    # circular reference
    from .client import Client

    # class is declared as generic in stubs but not at runtime
    AnyFuture = asyncio.Future[Any]
else:
    AnyFuture = asyncio.Future


@debugging
class Producer:
    """
    Produces messages
    """

    client: Client
    stream_name: str
    channel_key: str
    id: int

    log_debug: Callable[..., None]

    def __init__(self, client: Client, stream_name: str) -> None:
        """
        default constructor
        """
        Producer.log_debug("__init__ %r %r", client, stream_name)

        self.client = client
        self.stream_name = stream_name
        self.channel_key = "%s:responseid" % client.namespace
        self.id = time.time_ns()

    async def _resp_task(
        self, payload: Dict[str, Any], response_channel_id: str
    ) -> Any:
        """
        utility method for a confirmed request
        """
        if response_channel_id is None:
            # get a unique channel identifier
            uid = await self.client.redis.incr(self.channel_key)
            response_channel_id = "%s:response.%d" % (self.client.namespace, uid)
        Producer.log_debug("    - response_channel_id: %r", response_channel_id)

        response = {}
        try:
            # subscribe to the channel
            (response_channel,) = await self.client.redis.subscribe(response_channel_id)
            Producer.log_debug("    - response_channel: %r", response_channel)

            # pack it into the request payload
            payload["response_channel"] = response_channel_id

            # put the request into the stream
            message_id: str = await self.client.redis.xadd(self.stream_name, payload)
            Producer.log_debug("    - message_id: %r", message_id)

            # wait for the response to come back
            await response_channel.wait_message()
            try:
                response = await response_channel.get_json()
            except ValueError as err:
                response["message"] = "JSON Decoding Error"
                response["err"] = err

        except asyncio.CancelledError as err:
            Producer.log_debug("    - canceled %s", response_channel_id)
            response["message"] = "Cancelled Error"
            response["err"] = err
        finally:
            Producer.log_debug("    - finally %s", response_channel_id)
            await self.client.redis.unsubscribe(response_channel)

        return response

    # pylint: disable=invalid-name
    def addUnconfirmedMessage(
        self, message: Any, response_channel_id: str = None
    ) -> AnyFuture:
        """
        Return a task that adds an unconfirmed message to the message queue.
        """
        Producer.log_debug("addUnconfirmedMessage %r", message)

        # JSON encode the message
        payload = {"message": json.dumps(message)}
        if response_channel_id is not None:
            payload["response_channel"] = response_channel_id
        # create a task to add it to the stream
        future = self.client.redis.xadd(self.stream_name, payload)

        return cast(AnyFuture, future)

    # pylint: disable=invalid-name
    def addConfirmedMessage(
        self, message: Any, response_channel_id: str = None
    ) -> AnyFuture:
        """
        Return a task that adds a confirmed message to the message queue and
        waits for the response.
        """
        Producer.log_debug("addConfirmedMessage %r", message)

        # JSON encode the message
        payload = {"message": json.dumps(message)}

        # create a task to add it to the stream
        future = self._resp_task(payload, response_channel_id)

        return cast(AnyFuture, future)

    # pylint: disable=invalid-name
    def destroy( self ) -> None:
        """ destroy- stops this producer from working. This is automatically called when
            client.dispose_producer() is called with this producer
        """
        self.client = None
        self.stream_name = None
        Producer.log_debug("producer id %r destroyed", self.id)



"""
Client for RedisMQ
"""

from __future__ import annotations

import aioredis  # type: ignore

from typing import Any, Callable, Optional

from .debugging import debugging
from .producer import Producer
from .consumer import Consumer

__all__ = ["Client"]


@debugging
class Client:
    """
    Top level class to interface with message queue
    """

    log_debug: Callable[..., None]

    def __init__(self) -> None:
        """
        default constructor - use connect() instead
        """
        Client.log_debug("__init__")

        self.redis: Any = None
        self.namespace: str = "rmq"
        self.producer_registry: object = {}
        self.status: str = "wait"

    # this is a classmethod so it can be async
    @classmethod
    async def connect(cls, address: str, namespace: Optional[str] = None) -> "Client":
        """
        Call to create a connected client.
        """
        Client.log_debug("connect %s", address)
        client = cls()
        client.status = "connecting"
        client.redis = await aioredis.create_redis_pool(address, encoding="utf-8")
        if namespace:
            client.namespace = namespace
        client.status = "ready"
        return client

    async def close(self) -> None:
        """
        Call to close a connected client.
        """
        Client.log_debug("close")
        if self.status in ["closed", "closing"]:
            raise RuntimeError("Client.close() has already been called")
        if self.status != "ready":
            raise RuntimeError("Client is not ready to close")
        self.status = "closing"
        self.redis.close()
        await self.redis.wait_closed()
        self.status = "closed"

    async def producer(self, stream_name: str) -> Producer:
        """
        Use this to get a Producer
        """
        Client.log_debug("producer %s", stream_name)
        if stream_name not in self.producer_registry:
            Client.log_debug("    - adding producer %s to registry", stream_name)
            self.producer_registry[stream_name] = Producer(self, stream_name)
        else:
            Client.log_debug("    - producer %s found in registry", stream_name)
        return self.producer_registry[stream_name]

    async def dispose_producer(self, producer: Producer) -> None:
        """ dispose_producer of client
            @param producer - the producer to dispose of
        """
        stream_name = producer.stream_name
        if self.producer_registry[stream_name] is producer:
            del self.producer_registry[stream_name]
            producer.destroy()
            Client.log_debug("dispose_producer - producer %s disposed", stream_name)
        else:
            raise ValueError("Producer named %s registry entry mismatch." % stream_name)

    async def consumer(
        self,
        stream_name: str,
        group_name: str,
        consumer_id: str,
        # scan_pending_on_start: bool = True,
        # claim_stale_messages: bool = True,
        # min_idle_time: int = 60000,
    ) -> Consumer:
        """
        Use this to get a Consumer
        """
        Client.log_debug("consumer %s ...", stream_name)

        try:
            info = await self.redis.xinfo_groups(stream_name)
            Client.log_debug("    - xinfo_groups %s", info)
        except aioredis.RedisError as err:
            Client.log_debug("    - no existing stream %s", stream_name)
            info = []

        if not any(e["name"] == group_name for e in info):
            await self.redis.xgroup_create(
                stream_name, group_name, latest_id="$", mkstream=True
            )
            Client.log_debug(
                "    - added group %s to stream %s ", group_name, stream_name
            )

        return Consumer(
            self,
            stream_name,
            group_name,
            consumer_id,
            # scan_pending_on_start,
            # claim_stale_messages,
            # min_idle_time,
        )



# setup.py

import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

with open(HERE / "README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="redismq",
    version="1.0.6",
    author="Steve Mandl",
    author_email="sjm34@cornell.edu",
    description="Message Queueing for Redis Streams",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stevemandl/pyredismq",
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=['aioredis'],
    extras_require={
        'dev': ['pytest', 'pytest-asyncio', 'pytest-cov', 'pytest-timeout', 'pylint']
    },
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)
